#ifndef _OBJECT_COLLISION_H
#define _OBJECT_COLLISION_H

#include "types.h"

// extern ? debug_print_obj_collision(?);
// extern ? detect_object_hitbox_overlap(?);
// extern ? detect_object_hurtbox_overlap(?);
// extern ? clear_object_collision(?);
// extern ? check_collision_in_list(?);
// extern ? check_player_object_collision(?);
// extern ? check_pushable_object_collision(?);
// extern ? check_destructive_object_collision(?);
extern void detect_object_collisions(void);

#endif /* _OBJECT_COLLISION_H */
